#include "fap.h"
#include "main.h"

static void fap_transfer_nrf_fifo_to_sw_fifo(void);
static void fap_set_system_idle(void);

typedef struct
{
  uint8_t adr_device0[5];           // The FAP uses fixed 5 byte address  
  uint8_t adr_device1[5];           // The FAP uses fixed 5 byte address
  uint8_t adr_device2;
  uint8_t adr_device3;
  uint8_t adr_device4;
  uint8_t adr_device5;
  uint8_t pl_length_device0;
  uint8_t pl_length_device1;
  uint8_t pl_length_device2;
  uint8_t pl_length_device3;
  uint8_t pl_length_device4;
  uint8_t pl_length_device5;
  uint8_t fap_ch_tab[FAP_CH_TAB_SIZE];
}fap_dynamic_parameters_t;

static fap_dynamic_parameters_t fap_parameters=
{
  FAP_ADDRESS_DEVICE0,
  FAP_ADDRESS_DEVICE1,
  FAP_ADDRESS_DEVICE2,
  FAP_ADDRESS_DEVICE3,
  FAP_ADDRESS_DEVICE4,
  FAP_ADDRESS_DEVICE5,
  FAP_PL_LENGTH_DEVICE0,
  FAP_PL_LENGTH_DEVICE1,
  FAP_PL_LENGTH_DEVICE2,
  FAP_PL_LENGTH_DEVICE3,
  FAP_PL_LENGTH_DEVICE4,
  FAP_PL_LENGTH_DEVICE5,
  FAP_CHANNELS
};

// Volatile status variables and flags
static volatile bool fap_start_new_tx;
static volatile bool fap_tx_rx_infinite_timeout;
static volatile bool fap_lp_rx_enable;

static volatile uint8_t fap_radio_startup_counter;
static volatile int16_t fap_latency_counter;
static volatile uint8_t fap_retry_extend_counter;
static volatile uint16_t fap_try_counter;

static volatile uint16_t fap_timer_period;
static volatile uint8_t fap_sync_timer_period;

static volatile fap_modes_t fap_mode;
static volatile fap_tx_status_t fap_tx_status;
static volatile fap_rx_status_t fap_rx_status;

// Status variables
static uint8_t fap_ch_tab_idx_tx;
static uint8_t fap_ch_tab_idx_rx;
static uint8_t fap_selected_idle_mode;

// For providing transmission statistics only
static uint16_t fap_ch_switch_counter;

// SW FIFO
static uint8_t fap_sw_fifo_head, fap_sw_fifo_tail; 
static fap_rx_struct_t fap_sw_fifo[FAP_FIFO_SIZE];

//***************************************************************************** 
// Function bodies, application
//*****************************************************************************

void fap_init(void)
{
  FAP_NRF_IRQ_DISABLE();
  FAP_TIMER_IRQ_DISABLE();

  fap_mode=FAP_IDLE;

  CE_LOW();
  
  // Status variable init
  fap_ch_tab_idx_tx=0;
  fap_ch_tab_idx_rx=0;
  fap_sw_fifo_head=0; 
  fap_sw_fifo_tail=0; 
  fap_timer_period=0;
  fap_sync_timer_period=0;
  
  // Default setup that may be altered by application
  fap_selected_idle_mode=FAP_PDOWN_IDLE;  // Radio in power down between TX or RX      
  fap_lp_rx_enable=FALSE;                 // Uses low latency rx mode
  
  // Fixed parameter setup
  hal_nrf_set_datarate(HAL_NRF_2MBPS);    // 2Mbps datarate (maximum)  
  hal_nrf_set_auto_retr(FAP_NRF_AUTO_RETRIES, 500); // 500 us retransmit delay (minimum)   
  hal_nrf_set_output_power(HAL_NRF_0DBM); // 0 dBm output power (maximum)
  hal_nrf_set_crc_mode(HAL_NRF_CRC_16BIT);  // 16 bit CRC (maximum)
  
  hal_nrf_write_reg(DYNPD,0xff);
  hal_nrf_write_reg(FEATURE,0xff);
  
  // General radio setup
  hal_nrf_set_rf_channel(fap_parameters.fap_ch_tab[fap_ch_tab_idx_tx]);
  hal_nrf_get_clear_irq_flags();   

  hal_nrf_set_address(HAL_NRF_PIPE0, fap_parameters.adr_device0);
  hal_nrf_set_address(HAL_NRF_PIPE1, fap_parameters.adr_device1);
  hal_nrf_set_address(HAL_NRF_PIPE2, (uint8_t*)&fap_parameters.adr_device2);
  hal_nrf_set_address(HAL_NRF_PIPE3, (uint8_t*)&fap_parameters.adr_device3);
  hal_nrf_set_address(HAL_NRF_PIPE4, (uint8_t*)&fap_parameters.adr_device4);
  hal_nrf_set_address(HAL_NRF_PIPE5, (uint8_t*)&fap_parameters.adr_device5);

  hal_nrf_set_rx_pload_width(HAL_NRF_PIPE0, FAP_PL_LENGTH_DEVICE0);  
  hal_nrf_set_rx_pload_width(HAL_NRF_PIPE1, FAP_PL_LENGTH_DEVICE1);  
  hal_nrf_set_rx_pload_width(HAL_NRF_PIPE2, FAP_PL_LENGTH_DEVICE2);  
  hal_nrf_set_rx_pload_width(HAL_NRF_PIPE3, FAP_PL_LENGTH_DEVICE3);  
  hal_nrf_set_rx_pload_width(HAL_NRF_PIPE4, FAP_PL_LENGTH_DEVICE4);  
  hal_nrf_set_rx_pload_width(HAL_NRF_PIPE5, FAP_PL_LENGTH_DEVICE5);  

  fap_flush_rx_fifo();
  
  FAP_NRF_IRQ_ENABLE();
  FAP_TIMER_IRQ_ENABLE(); 
}  

void fap_select_rx_power_mode(fap_rx_modes_t mode)
{
  fap_lp_rx_enable=(bool)mode;
}

void fap_rx_data(uint8_t rx_setup, uint16_t rx_timeout)
{
  fap_goto_idle_mode();

  FAP_TIMER_IRQ_DISABLE();
  FAP_NRF_IRQ_DISABLE();
  hal_nrf_close_pipe(HAL_NRF_ALL);
 
  // Restore pipe 0 address 
  hal_nrf_set_address(HAL_NRF_PIPE0, fap_parameters.adr_device0);
  
  switch(rx_setup)
  {
    case FAP_DEVICE0:
      hal_nrf_open_pipe(HAL_NRF_PIPE0, TRUE);
      break;
    case FAP_DEVICE1:
      hal_nrf_open_pipe(HAL_NRF_PIPE1, TRUE);
      break;
    case FAP_DEVICE2:
      hal_nrf_open_pipe(HAL_NRF_PIPE2, TRUE);
      break;
    case FAP_DEVICE3:
      hal_nrf_open_pipe(HAL_NRF_PIPE3, TRUE);    
      break;
    case FAP_DEVICE4:
      hal_nrf_open_pipe(HAL_NRF_PIPE4, TRUE);    
      break;
    case FAP_DEVICE5:
      hal_nrf_open_pipe(HAL_NRF_PIPE5, TRUE);    
      break;
    case FAP_HOST:
      hal_nrf_open_pipe(HAL_NRF_ALL, TRUE);
      break;
  } 
  hal_nrf_set_power_mode(HAL_NRF_PWR_UP);
  hal_nrf_set_operation_mode(HAL_NRF_PRX);
      
  if(rx_timeout==0)
  {
    fap_tx_rx_infinite_timeout=TRUE;
  }
  else
  {
    fap_latency_counter=rx_timeout;
    fap_tx_rx_infinite_timeout=FALSE;
  }
  fap_timer_period=0;
  fap_mode=FAP_RECEIVING;
  
  FAP_NRF_IRQ_ENABLE();
  FAP_TIMER_IRQ_ENABLE();
}

void fap_tx_data(uint8_t *datainput, uint8_t tx_dest, uint16_t tx_timeout)
{
  uint8_t device_pl_length, device_byte_address;
  
  fap_goto_idle_mode();
  
  FAP_TIMER_IRQ_DISABLE();
  FAP_NRF_IRQ_DISABLE();

  hal_nrf_set_power_mode(HAL_NRF_PWR_UP);
  
  // Transmit address setup
  if(tx_dest==FAP_DEVICE0)
  {
    device_pl_length=fap_parameters.pl_length_device0;
    hal_nrf_set_address(HAL_NRF_TX, fap_parameters.adr_device0);
    hal_nrf_set_address(HAL_NRF_PIPE0, fap_parameters.adr_device0);
  }
  else
  { 
    hal_nrf_set_address(HAL_NRF_TX, fap_parameters.adr_device1);
    hal_nrf_set_address(HAL_NRF_PIPE0, fap_parameters.adr_device1);
        
    if(tx_dest!=FAP_DEVICE1)
    {
      switch(tx_dest)
      {
        default:
        case FAP_DEVICE2:
          device_byte_address=fap_parameters.adr_device2;
          device_pl_length=fap_parameters.pl_length_device2;
          break;
        case FAP_DEVICE3:
          device_byte_address=fap_parameters.adr_device3;
          device_pl_length=fap_parameters.pl_length_device3;
          break;
        case FAP_DEVICE4:
          device_byte_address=fap_parameters.adr_device4;
          device_pl_length=fap_parameters.pl_length_device4;
          break;
        case FAP_DEVICE5:
          device_byte_address=fap_parameters.adr_device5;
          device_pl_length=fap_parameters.pl_length_device5;
          break;
      }
      hal_nrf_write_reg(RX_ADDR_P0, device_byte_address);
      hal_nrf_write_reg(TX_ADDR, device_byte_address);
    }
    else
    {
      device_pl_length=fap_parameters.pl_length_device1; 
    }
  }

  hal_nrf_open_pipe(HAL_NRF_PIPE0, TRUE);
  hal_nrf_set_operation_mode(HAL_NRF_PTX);
  fap_ch_switch_counter=0; 
  fap_retry_extend_counter=0;
  fap_try_counter=0;

  if(tx_timeout==0)
  {
    fap_tx_rx_infinite_timeout=TRUE;
  }
  else
  {
    fap_latency_counter=tx_timeout;
    fap_tx_rx_infinite_timeout=FALSE;
  }
  
  hal_nrf_flush_tx();
  hal_nrf_write_tx_pload(datainput, device_pl_length);
  hal_nrf_set_rf_channel(fap_parameters.fap_ch_tab[fap_ch_tab_idx_tx]);
  
  if(fap_selected_idle_mode==FAP_PDOWN_IDLE)
  {
    fap_radio_startup_counter=FAP_RADIO_PWR_UP_DELAY;
  }
  else
  {
    fap_radio_startup_counter=0; 
  }
 
  fap_start_new_tx=TRUE;
  fap_tx_status=FAP_TX_FAILURE; 
  fap_mode=FAP_TRANSMITTING;

  FAP_NRF_IRQ_ENABLE();
  FAP_TIMER_IRQ_ENABLE();
}

fap_modes_t fap_get_mode(void)
{
  return fap_mode;
}

fap_rx_status_t fap_get_rx_fifo_status(void)
{
  return fap_rx_status; 
}

bool fap_read_rx_fifo(fap_rx_struct_t* return_struct)
{
  uint8_t index;

  if(fap_rx_status!=FAP_RX_FIFO_EMPTY)
  {
    FAP_NRF_IRQ_DISABLE();
    FAP_TIMER_IRQ_DISABLE();

    for(index=0; index<fap_sw_fifo[fap_sw_fifo_tail].rx_pl_length; index++)
    {
      return_struct->rx_pl[index]=fap_sw_fifo[fap_sw_fifo_tail].rx_pl[index];
    }
    return_struct->rx_pl_length=fap_sw_fifo[fap_sw_fifo_tail].rx_pl_length;
    return_struct->rx_pipe=fap_sw_fifo[fap_sw_fifo_tail].rx_pipe;

    fap_sw_fifo_tail=(fap_sw_fifo_tail+1)%FAP_FIFO_SIZE;
    if(fap_sw_fifo_tail==fap_sw_fifo_head)
    {
      fap_rx_status=FAP_RX_FIFO_EMPTY;
    }
    
    fap_transfer_nrf_fifo_to_sw_fifo();

    FAP_NRF_IRQ_ENABLE();
    FAP_TIMER_IRQ_ENABLE();
       
    return TRUE;
  }  
  return FALSE;
}

fap_tx_status_t fap_get_tx_status(void)
{
   return fap_tx_status;   
}

void fap_select_radio_idle_mode(fap_radio_idle_modes_t mode)
{
  FAP_NRF_IRQ_DISABLE();
  FAP_TIMER_IRQ_DISABLE();
 
  fap_selected_idle_mode=mode;
  // Update radio power mode if FAP is in idle mode
  if(fap_mode==FAP_IDLE)
  {
    hal_nrf_set_power_mode(fap_selected_idle_mode);
  }
 
  FAP_NRF_IRQ_ENABLE();
  FAP_TIMER_IRQ_ENABLE();
}

uint16_t fap_get_retries(void)
{
  return fap_try_counter;
}

uint16_t fap_get_ch_switches(void)
{
  return  fap_ch_switch_counter;
}

void fap_nrf_isr_function(void)
{
  uint8_t irq_flags;
  uint8_t tmp_loop=10;
  
  static uint8_t ch_swap_temp;
  static uint8_t ch_swap_idx;
  
  irq_flags=hal_nrf_get_clear_irq_flags();
  if(fap_mode==FAP_TRANSMITTING)
  {
    switch(irq_flags)
    {
      case (1<<TX_DS): 
        // Get number of tries. Add 1 for 1st attempt.
        fap_try_counter+=(hal_nrf_get_transmit_attempts()+1);  
        
        // Get "random" number between 0 and FAP_CH_TAB_SIZE-1
        // Used for random channel swap
        ch_swap_idx=(ch_swap_idx+fap_try_counter)%FAP_CH_TAB_SIZE;
        
        if(fap_try_counter>2)
        {
          fap_sync_timer_period=(fap_sync_timer_period+1)%FAP_CH_TAB_SIZE;
        }
        fap_tx_status=FAP_TX_SUCCESS;
        fap_set_system_idle();
        break; 

      case (1<<MAX_RT):
        fap_try_counter+=(FAP_NRF_AUTO_RETRIES+1);
        fap_retry_extend_counter=(fap_retry_extend_counter+1)%FAP_AUTO_RETRIES_EXTEND;
    
        if(fap_retry_extend_counter==0)
        {
          // Possible channel sync with other devices
          // Modify freqeuncy search order randomly
          if((fap_ch_switch_counter+1)%FAP_CH_TAB_SIZE==0)
          {
            ch_swap_temp=fap_parameters.fap_ch_tab[ch_swap_idx];
            fap_parameters.fap_ch_tab[ch_swap_idx]=fap_parameters.fap_ch_tab[fap_ch_tab_idx_tx];
            fap_parameters.fap_ch_tab[fap_ch_tab_idx_tx]=ch_swap_temp;
          }
          
          fap_ch_tab_idx_tx=(fap_ch_tab_idx_tx+1)%FAP_CH_TAB_SIZE;
          hal_nrf_set_rf_channel(fap_parameters.fap_ch_tab[fap_ch_tab_idx_tx]);
          fap_ch_switch_counter++;
        }

        fap_latency_counter-=(FAP_NRF_AUTO_RETRIES+1); 
        if(fap_latency_counter<1 && fap_tx_rx_infinite_timeout==FALSE)
        {
          fap_tx_status=FAP_TX_FAILURE;
          fap_set_system_idle();
        }
        else
        {
          //CE_PULSE(); // Start new tx on the new channel
			 CE_HIGH();
			 while(tmp_loop--)
			 {
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
			 }
			 CE_LOW();
        }
        break; 
    }
  }
  else
  {
  	if(fap_mode==FAP_RECEIVING)
  	{          
    	if(irq_flags==(1<<RX_DR))
    	{
		fap_transfer_nrf_fifo_to_sw_fifo();
		
    	}
  	}
  }
}

void fap_timer_isr_function(void)
{
  	uint8_t tmp_loop=10;
  if(fap_mode==FAP_TRANSMITTING) 
  { 
    if(fap_radio_startup_counter>0)
    {
      fap_radio_startup_counter--;
    }
    if((fap_start_new_tx==TRUE )&& (fap_sync_timer_period==fap_timer_period) && (fap_radio_startup_counter==0))
    {
      fap_start_new_tx=FALSE;
      //CE_PULSE();
		CE_HIGH();
		while(tmp_loop--)
		{
		  		__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
				__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
            __asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");__asm("NOP");
		}
		CE_LOW();
    }     
  }
  else
  if(fap_mode==FAP_RECEIVING)
  { 
    if(fap_tx_rx_infinite_timeout==TRUE || (fap_tx_rx_infinite_timeout==FALSE && fap_latency_counter>0))
    {
      fap_latency_counter--;
      if(fap_lp_rx_enable==FALSE)
      {
        CE_LOW();
        fap_ch_tab_idx_rx=(fap_ch_tab_idx_rx+1)%FAP_CH_TAB_SIZE;
        hal_nrf_set_rf_channel(fap_parameters.fap_ch_tab[fap_ch_tab_idx_rx]);
        CE_HIGH();
      } 
      else
      {
        if(fap_timer_period==0)
        {
          // Change channel
          if(fap_timer_period==0) 
          {
            fap_ch_tab_idx_rx=(fap_ch_tab_idx_rx+1)%FAP_CH_TAB_SIZE;
            hal_nrf_set_rf_channel(fap_parameters.fap_ch_tab[fap_ch_tab_idx_rx]);
          }

          hal_nrf_set_power_mode(HAL_NRF_PWR_UP);
        }

        if(fap_timer_period==FAP_RX_FULL_CH_REV)
        {
          hal_nrf_set_power_mode(HAL_NRF_PWR_DOWN);
        }

        if(fap_timer_period < FAP_RX_FULL_CH_REV)
        {
          // RX on
          if((fap_timer_period-FAP_RADIO_PWR_UP_DELAY)%FAP_RX_SINGLE_CH_REV==0)
          {
            CE_HIGH();    
          }
          // RX off  
          if((fap_timer_period-FAP_RADIO_PWR_UP_DELAY-FAP_LP_RX_LISTEN_PERIODS)%FAP_RX_SINGLE_CH_REV==0) 
          {
            CE_LOW();
          }
        }
      }
    }
    else
    {
      fap_set_system_idle();  
    }
  } 
 
  // Increment period counter
  if(fap_mode==FAP_RECEIVING && fap_lp_rx_enable==TRUE)
  {
    if(fap_timer_period<FAP_LP_RX_POLL_PERIOD)
    {
      fap_timer_period++;  
    }
    else
    {
      fap_timer_period=0;
    }
  }
  else
  {
    fap_timer_period=(fap_timer_period+1)%(FAP_CH_TAB_SIZE);
  }
}

bool fap_flush_rx_fifo(void)
{
  if(fap_mode==FAP_IDLE)
  {
    FAP_TIMER_IRQ_DISABLE(); 
    FAP_NRF_IRQ_DISABLE(); 

    hal_nrf_flush_rx();
    fap_sw_fifo_head=fap_sw_fifo_tail=0;
    fap_rx_status=FAP_RX_FIFO_EMPTY;
  
    FAP_TIMER_IRQ_ENABLE(); 
    FAP_NRF_IRQ_ENABLE();
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}

void fap_goto_idle_mode()
{
  fap_tx_rx_infinite_timeout=FALSE;
  fap_latency_counter=0; 
  // Wait for protocol to finish ongoing operations
  while(fap_mode!=FAP_IDLE);
}
//***************************************************************************** 
// Function bodies, internal
//*****************************************************************************

static void fap_set_system_idle(void)
{ 
  CE_LOW();
  fap_mode=FAP_IDLE;
  if(fap_selected_idle_mode==FAP_PDOWN_IDLE)
  {
    hal_nrf_set_power_mode(HAL_NRF_PWR_DOWN);
  }  
}

static void fap_transfer_nrf_fifo_to_sw_fifo(void)
{
  static uint16_t tempvalue;
 
  while(hal_nrf_rx_fifo_empty()==FALSE && fap_rx_status!=FAP_RX_FIFO_FULL)
  {
    tempvalue=hal_nrf_read_rx_pload(fap_sw_fifo[fap_sw_fifo_head].rx_pl); 
    fap_sw_fifo[fap_sw_fifo_head].rx_pipe=(uint8_t)(tempvalue>>8); // MSByte, pipe
    fap_sw_fifo[fap_sw_fifo_head].rx_pl_length=(uint8_t)(tempvalue); // LSByte=length
    fap_sw_fifo_head=(fap_sw_fifo_head+1)%FAP_FIFO_SIZE;
    
    if(fap_sw_fifo_head==fap_sw_fifo_tail)
    {
      fap_rx_status=FAP_RX_FIFO_FULL;
    }
    else
    {
      fap_rx_status=FAP_RX_DATA_IN_FIFO;
    }
  }
}

/**
@}
*/
