#ifndef	__FAP_H__
#define  __FAP_H__

#include "fap_setup.h"
#include "nRF24L01.h"

#define FAP_LP_RX_LISTEN_PERIODS 1                                           

//***************************************************************************** 
// Derived parameters
//*****************************************************************************

/**
Specifies number of transmit periods the transmitter resides at one channel before
channel switch. For the length of one transmit period in [us], please see the 
FAP documentation. This period equals the duration between the start of each 
reatransmit, and is derived form the retransmit delay and "on air" packet length.
The maximum length of one transmit period in [us] is specified by FAP_TX_PERIOD_MAX.
*/
#define FAP_RADIO_PWR_UP_DELAY ((2000/FAP_RX_PERIOD)+1)   // 2 ms radio startup

/**
Specifies number of transmit periods the transmitter resides at one channel before
channel switch. For the length of one transmit period in [us], please see the 
FAP documentation. This period equals the duration between the start of each 
reatransmit, and is derived form the retransmit delay and "on air" packet length.
The maximum length of one transmit period in [us] is specified by #FAP_MAX_TX_PERIOD.
*/
#define FAP_TX_SINGLE_CH_REV ((FAP_NRF_AUTO_RETRIES+1)*FAP_AUTO_RETRIES_EXTEND)//[#]  

/*
Specifies the maximum time in [us] the transmitter resides at a single channel before 
channel switch. This parameter is derived from the device maximum payload length.
*/  
#define FAP_TX_SINGLE_CH_REV_TIME (FAP_TX_SINGLE_CH_REV*FAP_MAX_TX_PERIOD)//[us]

/*
Specifies the number of transmit periods necessary  for a full revolution
through all channels. It is recommended that the transmit timeout period 
specified when using fap_tx_data() is equal or greater than this value.
The maximum length of one transmit period in [us] is specified by #FAP_MAX_TX_PERIOD.
*/
#define FAP_TX_FULL_CH_REV (FAP_TX_SINGLE_CH_REV*FAP_CH_TAB_SIZE)//[#]

/*
Specifies the time in [us] necacarry for the transmitter to revolve through 
all channels.
*/
#define FAP_TX_FULL_CH_REV_TIME (FAP_TX_FULL_CH_REV*FAP_MAX_TX_PERIOD)//[us]

/**
Specifies the number of recive periods that equals the time the transmitter
resides on a single channel before channel switch. 
The length of one tranmit period in [us] is specified by FAP_RX_PERIOD.
*/
#define FAP_RX_SINGLE_CH_REV ((uint16_t)(((FAP_TX_SINGLE_CH_REV*FAP_MAX_TX_PERIOD)/FAP_RX_PERIOD)+1))//[#]

/**
Specifies the time in [us] that equals the time the transmitter
resides on a single channel before channel switch.
Thus, FAP_RX_SINGLE_CH_REV_TIME=FAP_TX_SINGLE_CH_REV_TIME.      
*/
#define FAP_RX_SINGLE_CH_REV_TIME (FAP_RX_SINGLE_CH_REV*FAP_RX_PERIOD)//[us]

/**
Specifies the number of recive periods that equals the time the transmitter
uses on a full revolution through all channels. The length of one transmit 
period in [us] is specified by FAP_RX_PERIOD.     
*/
#define FAP_RX_FULL_CH_REV (FAP_RX_SINGLE_CH_REV*FAP_CH_TAB_SIZE) // [#]

/**
Specifies the time in [us] that equals the time the transmitter uses on 
a full revolution through all channels. Thus
FAP_RX_FULL_CH_REV=FAP_TX_FULL_CH_REV.    
*/
#define FAP_RX_FULL_CH_REV_TIME (FAP_RX_FULL_CH_REV*FAP_RX_PERIOD) // [us]

//***************************************************************************** 
// Typedefs
//*****************************************************************************

/**
Type used in RX mode for storing received data, datalength and datasource   
*/
typedef struct
{
  uint8_t rx_pl[FAP_MAX_PL_LENGTH];  
  uint8_t rx_pl_length;
  uint8_t rx_pipe;
} fap_rx_struct_t;

/**
Possible FAP modes. Return values for function fap_get_mode().    
*/
typedef enum
{
  FAP_IDLE,
  FAP_TRANSMITTING,
  FAP_RECEIVING
}fap_modes_t;

/**
Possible RX fifo states. Return values for function fap_get_rx_fifo_status().    
*/
typedef enum
{
  FAP_RX_FIFO_EMPTY,
  FAP_RX_DATA_IN_FIFO,
  FAP_RX_FIFO_FULL,
}fap_rx_status_t;

/**
Possible TX results. Return values for function fap_get_tx_status().    
*/
typedef enum
{
  FAP_TX_FAILURE,
  FAP_TX_SUCCESS,            
}fap_tx_status_t;

/**
Possible radio modes during FAP IDLE mode. Input paramerters for 
function fap_select_radio_idle_mode().    
*/
typedef enum
{
  FAP_PDOWN_IDLE,
  FAP_STANDBY_IDLE 
}fap_radio_idle_modes_t;

/**
Posseible receive modes. Input paramerters for 
function fap_select_rx_power_mode().    
*/
typedef enum
{
  FAP_LOW_LATENCY_RX,
  FAP_LOW_POWER_RX,  
}fap_rx_modes_t;

#define FAP_DEVICE0 HAL_NRF_PIPE0
#define FAP_DEVICE1 HAL_NRF_PIPE1
#define FAP_DEVICE2 HAL_NRF_PIPE2
#define FAP_DEVICE3 HAL_NRF_PIPE3
#define FAP_DEVICE4 HAL_NRF_PIPE4
#define FAP_DEVICE5 HAL_NRF_PIPE5
#define FAP_HOST 0xFF

//***************************************************************************** 
// Function prototypes and function parameter defines
//*****************************************************************************

/**
@name General functions
*/

/**
Initialization function for the frequency agility protocol (FAP). This function
must be called at least once befor any of the remaining FAP functions may be used. 
It is recommended that the FAP timer interrupt service routine is not 
enabeled until after the FAP is initialized.
*/
void fap_init(void);

/**
Function for selecting in which mode the radio shall reside when
the FAP is in IDLE mode. The nRF radio may reside in two different modes 
when not transmitting or receiving (FAP IDLE), 
which are STANDBY or POWER DOWN, respectively.
In STANDBY the radio consumes more power than in POWER DOWN, but 
the radio startup time is shorter, which will result in a lower transmit 
latency when starting a new transmit operation. This function selects
the radio mode to be used in FAP IDLE mode. 
 
@param mode selects the radio mode.    
Possible arguments are: 
@arg @c FAP_PDOWN_IDLE   
@arg @c FAP_STANDBY_IDLE 
*/
void fap_select_radio_idle_mode(fap_radio_idle_modes_t mode); // Power down / standby

/**
Function for forcing the FAP to IDLE mode. 
The FAP have three different modes; TRANSMITTING,
RECEIVING or IDLE, respectively. When the FAP is receiving 
or transmitting it may automatically return to IDLE mode after 
after a specified or in TRANSMIT mode also after tranmission
success. In addition, fap_goto_idle_mode() may
be used in order to immediate force the FAP to IDLE mode.
*/
void fap_goto_idle_mode(void);

/**
Function for observing the current FAP mode.

@return 
Returns the current mode for the FAP.
@retval FAP_IDLE
@retval FAP_TRANSMITTING 
@retval FAP_RECEIVING

@sa fap_rx_data(), fap_tx_data(), fap_goto_idle_mode()
*/
fap_modes_t fap_get_mode(void);  

/** @} */

/**
@name Transmit related functions
*/ 

/**
Function for transmitting data. This function transmits the data provided in
*datainput to the destination given by tx_dest. When the data is successfully 
transmitted the FAP will automtically return to IDLE mode. It may be specified a maximum 
period the FAP shall try to transmit data before returning to IDLE mode. 

A typical usage of this function is to first start the transmit operation using
fap_tx_data(), then waiting for the FAP to automatically return to
IDLE mode while checking fap_get_mode() and finally getting the result
of the transmit operation using fap_get_tx_status().   

@param *datainput Source buffer for the data to be transmitted. 

@param tx_dest is the destination address for the transmission.    
Possible arguments are: 
@arg @c FAP_DEVICE0 for transmiting to address 0   
@arg @c FAP_DEVICE1 for transmitting to address 1  
@arg @c FAP_DEVICE2 for transmitting to address 2   
@arg @c FAP_DEVICE3 for transmitting to address 3   
@arg @c FAP_DEVICE4 for transmitting to address 4   
@arg @c FAP_DEVICE5 for transmitting to address 5   

@param transmit_timeout specifies the maximum period the FAP shall attempt to
transmit before returning to IDLE mode. The maximum timeout period will 
be equivalent to transmit_timeout*FAP_MAX_TX_PERIOD [us]. An infinite transmit 
period may be specified by setting this parameter to 0. In any case the FAP
may be forced to IDLE mode by using the function fap_goto_idle_mode().
If a finite timeout is specified, it is recommended that the timeout value 
is set equal or grater than #FAP_TX_FULL_CH_REV.
    
@sa fap_get_mode(), fap_get_tx_status(), fap_goto_idle_mode(void)
*/
void fap_tx_data(uint8_t *datainput, uint8_t tx_dest, uint16_t tx_timeout);

/**
Function for getting the result of the previous transmit operation. 

@return
Result of the previous transmit operation.  
@retval FAP_TX_FAILURE if data were not successfully transmitted   
@retval FAP_TX_SUCCESS if data were successfully transmitted
@sa fap_tx_data(), fap_get_mode(), fap_goto_idle_mode(void)
*/
fap_tx_status_t fap_get_tx_status(void); 

/**
Function for providing the number of transmission
attempts needed during the previous transmit session. 
This function may be used as an indication on
the current radio transmit conditions. In a noisy 
environment one will experience an increase in the 
number of transmission attempts needed for a 
succesfull packet delivery to the receiver. Note that 
even in an environment without any interfering radio
sources one will experience that the FAP sometimes
uses several transmission attempts. This due to the
frequency altering scheme used at the receiver.      

@sa fap_get_ch_switches()
*/
uint16_t fap_get_retries(void);

/**
Function for providing the number of 
frequency channel switches neede during the previous 
transmit session. In an environment without any 
interfering radio sources one will not normally 
experience that the FAP needs to switch transmission 
frequency. A large number of channel switches
will be an indication on extensive radio 
interference.  

@sa fap_get_retries()
*/
uint16_t fap_get_ch_switches(void);

/** @} */

/** @name Receive related functions */
/**
Function for setting the FAP in receive mode and start monitoring the air for data. 
Received data will be stored in a FIFO which can be read out by the function
fap_read_rx_fifo(). It may be specified a timeout period after which 
the FAP shall automatically stop receiving and return to IDLE mode.    

@param rx_setup selects address(es) to monitor for incomming data.
Possible arguments are: 
@arg @c FAP_DEVICE0 for receiving at address 0 only  
@arg @c FAP_DEVICE1 for receiving at address 1 only  
@arg @c FAP_DEVICE2 for receiving at address 2 only  
@arg @c FAP_DEVICE3 for receiving at address 3 only  
@arg @c FAP_DEVICE4 for receiving at address 4 only  
@arg @c FAP_DEVICE5 for receiving at address 5 only  
@arg @c FAP_HOST for receiving at all addresses   

@param receive_timeout specifies the receive timeout period. The timeout period
will be equivalent to receive_timeout*FAP_RX_PERIOD [us]. By setting this 
parameter to 0 the receiver will remain in receive mode until it
is forced to IDLE mode by fap_goto_idle_mode(). If a finite timeout
is specified, it is recommended that the timeout value is set equal or 
grater than FAP_RX_FULL_CH_REV.

@sa fap_get_mode(), fap_get_rx_fifo_status(), fap_read_rx_fifo(), fap_select_rx_power_mode, 
fap_goto_idle_mode()
*/
void fap_rx_data(uint8_t rx_setup, uint16_t receive_timeout); 

/**
Function for setting the FAP receive power mode. The FAP offers
two different receive modes; one low power mode and one low latency mode.
In low power mode the FAP is monitoring the air for data in only short
slots of time and the receiver is switched off most of the time. 
This leads to a low power consumption at the receive side but a higher latency
experienced at the transmit side as the transmitter has to "struggle"
more to hit the the time slots where the receiver is listening. In low
latency mode the receiver is continously monitoring the air for data, 
which gives a higher power consumption at the receive side but a lower 
experienced latency at the transmit side.
The FAP do not have to be in IDLE mode before using this function. If 
the FAP is in receive mode when this function is used the new 
receive settings will apply immediately.  

@param mode selects the new receive mode.

Possible arguments are: 
@arg @c FAP_LOW_LATENCY_RX,
@arg @c FAP_LOW_POWER_RX, 

@sa fap_rx_data()
*/
void fap_select_rx_power_mode(fap_rx_modes_t mode);

/**
Function for getting the status of the receive FIFO. 

@return 
Returns the receive FIFO status.
@retval FAP_RX_FIFO_EMPTY if no unread data in receive FIFO
@retval FAP_RX_DATA_IN_FIFO if receive FIFO contains unread data  
@retval FAP_RX_FIFO_FULL if receive FIFO is full. In this state no new
data can be received until the FIFO is read or cleared. The FIFO may be
cleared by reading the FIFO contents using fap_read_rx_fifo() or 
by using fap_flush_rx_fifo().

@sa fap_rx_data(), fap_get_rx_fifo_status(), fap_read_rx_fifo(), fap_flush_rx_fifo(), 
fap_select_rx_power_mode(), fap_get_mode()
*/
fap_rx_status_t fap_get_rx_fifo_status(); 

/**
Function for flushing the FAP receive FIFO consisting of the nRF radio hardware
FIFO and a software FIFO. For this function to have any effect the FAP must
reside in IDLE mode when the function is called.

@retval FALSE if FIFO not flushed due to the FAP not being in IDLE mode.
@retval TRUE if FIFO flushed.
*/
bool fap_flush_rx_fifo(void);

/**
Function for reading received data from receive FIFO.
When the FAP is in recive mode all incoming data will
be stored in a FIFO which can be read using this function.

@param *fap_rx_struct specifies destination to where FIFO 
read data shall be copied. 

@return 
Information on whether FIFO contained data or not.  
@retval TRUE if unread FIFO data were copied to destination.
@retval FALSE if FIFO was empty and no data were copied.  
@sa fap_rx_data(), fap_get_rx_fifo_status(), 
*/
bool fap_read_rx_fifo(fap_rx_struct_t* return_struct);

/** @} */

/** @name Interrup service routine functions */

/**
Function to be called by the nRF radio interrupt service routine.  
*/
void fap_nrf_isr_function(void);

/**
Function to be calles by the FAP timer interrupt service routine.
The FAP requires one dedicated auto reload timer. This function must
be called by the FAP timer interrupt service routine. 
*/
void fap_timer_isr_function(void);

fap_rx_status_t fap_get_rx_fifo_status(void);
/**
@}
*/


#endif
